import os
import sys
#import sqlite3
#import datetime
#import pytz
#import copy
import itertools
import operator
import csv
import zipfile
import tempfile
import logging

import pandas as pd

from sqlalchemy import inspect
from sqlalchemy import create_engine, MetaData
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql import select, insert
from sqlalchemy.schema import CreateTable

from sqlalchemy.dialects.postgresql import insert as postgresql_insert

def createEngineFromFilename (sql_filename, echo = False):

	return create_engine("sqlite:///%s" % sql_filename, echo = echo)

def createEngineFromAddress (sql_address, echo = False):

	return create_engine("postgresql://%s" % sql_address, echo = echo)

def createEngineFromConfig (config_data, echo = False):

	if config_data.type == 'sqlite': return create_engine("sqlite:///%s" % config_data.filename, echo = echo)
	else: return create_engine("postgresql://%s" % config_data.sql_address, echo = echo)

def startSessionFromConfig (config_data, echo = False):

	sql_engine = createEngineFromConfig(config_data, echo)
	Session = sessionmaker(bind=sql_engine)
	return Session()
	
def startSessionFromEngine (engine):
	
	# Start the session
	Session = sessionmaker(bind=engine)
	return Session()

def createAllFromConfig (config_data, engine):

	# Create the tables
	config_data.meta.create_all(engine)

def printCreateSQL (sql_tables):

	# Loop tables and print SQL statments
	for sql_table in sql_tables:
		print(CreateTable(sql_table))

def foreignKeyPairs (sql_tables):

	# Assign the pairs of foreign and parent keys
	for fk_sql_table in sql_tables:
		for column in fk_sql_table.columns:
			if column.foreign_keys:
				for foreign_key in column.foreign_keys:
					yield foreign_key.column, column

def prepDataFrameUsingConfig (config_data, table, dataframe):

	# Confirm the table is within the config file
	if not table in config_data: raise Exception(f'{table} not found in database')

	col_header_dict = config_data._table_col_to_label[table]
	col_label_dict = config_data._table_label_to_col[table]

	# Check which header values were found
	col_header_count = len(set(dataframe.columns) & set(list(col_header_dict)))
	col_label_count = len(set(dataframe.columns) & set(list(col_label_dict)))

	# Fix the headers, if necessary 
	header_method = 'header' if col_header_count > col_label_count else 'label'
	if header_method == 'label': dataframe = dataframe.rename(columns = col_label_dict)

	# Define the standard columns
	std_cols = list(col_header_dict)

	# Check if the file has non-standard columns
	non_std_cols = list(set(dataframe.columns) - set(std_cols))
	dataframe = dataframe.drop(columns = non_std_cols)

	# Return the dataframe
	return dataframe

class SQLSelect ():
	def __init__ (self, sql_select_tables = [], sql_select_columns = [], sql_where = [], sql_tables = [], tables_in_select = [], sql_connection = None):
		self.select_results = None
		self._sql_select_tables = sql_select_tables
		self._sql_select_columns = sql_select_columns
		self._sql_where = sql_where
		self._sql_tables = sql_tables
		self._tables_in_select = tables_in_select 
		self._sql_connection = sql_connection

	@property
	def _total_tables_in_select (self):

		return sum(self._tables_in_select)

	@property
	def _sql_tables_in_select (self):

		return [sql_table for table_bool, sql_table in zip(self._tables_in_select, self._sql_tables) if table_bool]

	def select (self):
		if self._total_tables_in_select > 1:
			for parent_key, foreign_key in foreignKeyPairs(self._sql_tables_in_select):
				if not self._sql_where:
					self._sql_where = SQLWhere.fromQuery(self._sql_tables)
				self._sql_where.addEQColWhere(parent_key, foreign_key)
		self.select_results = list(self._sql_connection.execute(select(self.removeRepeatsInSelect(self._sql_select_tables, self._sql_select_columns)).where(self._sql_where.where_statement)))

	def toDataFrame (self):

		# Return as dataframe
		return(pd.DataFrame([dict(_r) for _r in self.select_results]))

	def toFile (self, out_filename, sep, warn_if_nothing = True):

		# Check if any data was returned, if not return nothing
		if len(self.select_results) == 0:
			if warn_if_nothing: logging.warning(f'Nothing to return. File ({out_filename}) not created')
			else: logging.info(f'Nothing to return.')
			return

		# Create the output file, using DictWriter
		header = self.select_results[0].keys()
		with open(out_filename, 'w') as entries_file:
			entries_writer = csv.DictWriter(entries_file, fieldnames = header, delimiter = sep)
			entries_writer.writeheader()
			for results_row in self.select_results:
				entries_writer.writerow(dict(results_row))

		# Update log
		logging.info('Retrieved results written to file (%s)' % out_filename)

	def toScreen (self, sep, warn_if_nothing = True):

		# Check if any data was returned, if not return nothing
		if len(self.select_results) == 0:
			if warn_if_nothing: logging.warning(f'Nothing to return.')
			else: logging.info(f'Nothing to return.')
			return

		# Print the header using sep
		header = self.select_results[0].keys()
		print(sep.join(header))

		# Print the values to stdout, seperated by sep
		for results_row in self.select_results:
			print(sep.join(map(str, results_row.values())))

		# Update log
		logging.info('Retrieved results sent to stdout')

	@classmethod
	def fromConfig (cls, config_data, sql_connection):

		# Create a SQLQuery object from the config file and sql_connection
		return cls(sql_select_tables = [], 
				   sql_select_columns = [], 
				   sql_where = [], 
				   sql_tables = config_data.sql_tables, 
				   tables_in_select = [False] * len(config_data.sql_tables), 
				   sql_connection = sql_connection)

	def addTableToSelect (self, table):

		# Add the table, update the tables in use
		for table_pos, sql_table in enumerate(self._sql_tables):
			if table == sql_table:
				self._tables_in_select[table_pos] = True
		if table in self._sql_select_tables: raise Exception('Table (%s) already assigned' % table)
		self._sql_select_tables.append(table)

	def addTablesToSelect (self, tables):

		# Add the tables, update the tables in use
		for table in tables:
			for table_pos, sql_table in enumerate(self._sql_tables):
				if table == sql_table:
					self._tables_in_select[table_pos] = True
			if table in self._sql_select_tables: raise Exception('Table (%s) already assigned' % table)
			self._sql_select_tables.append(table)

	def addColumnToSelect (self, column):

		# Add the column, update the tables in use
		for table_pos, sql_table in enumerate(self._sql_tables):
			for sql_column in sql_table.columns:
				if column == sql_column:
					self._tables_in_select[table_pos] = True
		if column in self._sql_select_columns: raise Exception('Column (%s) already assigned' % column)
		self._sql_select_columns.append(column)

	def addColumnsToSelect (self, columns):

		# Add the columns, update the tables in use
		for column in columns:
			for table_pos, sql_table in enumerate(self._sql_tables):
				for sql_column in sql_table.columns:
					if column == sql_column:
						self._tables_in_select[table_pos] = True
			if column in self._sql_select_columns: raise Exception('Column (%s) already assigned' % column)
			self._sql_select_columns.append(column)

	def addDictWhere (self, where_dict, include = None, cmp_type = None, dict_type = None):

		if not include: raise Exception('Must indicate include/exclude type for where statement')
		if not cmp_type: raise Exception('Must indicate comparison type for where statement')
		if not dict_type: raise Exception('Must indicate dict type for where statement')

		# Add the where statement from a dict, update the tables in use
		if not self._sql_where:
			self._sql_where = SQLWhere.fromQuery(self._sql_tables)
		if dict_type == 'Column': self._sql_where.addColDictWhere(where_dict, include = include, cmp_type = cmp_type)
		elif dict_type == 'Table': self._sql_where.addTableDictWhere(where_dict, include = include, cmp_type = cmp_type)
		for table_pos in self._sql_where._tables_used:
			if not self._tables_in_select[table_pos]:
				self._tables_in_select[table_pos] = True

	@staticmethod
	def removeRepeatsInSelect (sql_tables, sql_columns):

		remove_dict = {}
		for sql_table in sql_tables:
			for column in sql_table.columns:
				if column.foreign_keys:
					for foreign_key in column.foreign_keys:
						if foreign_key.column.table in sql_tables:
							remove_dict[sql_table] = column

		updated_tables = []
		updated_columns = []
		for sql_table in sql_tables:
			if sql_table not in remove_dict:
				updated_tables.append(sql_table)
			else:
				for column in sql_table.columns:
					if column != remove_dict[sql_table]:
						updated_columns.append(column)

		all_columns = list(itertools.chain.from_iterable([list(updated_table.columns) for updated_table in updated_tables])) + updated_columns
		for sql_column in sql_columns:
			if sql_column not in all_columns:
				updated_columns.append(sql_column)

		return updated_tables + updated_columns

class SQLUpdate ():
	def __init__ (self, sql_update = None, sql_values = {}, sql_where = [], sql_tables = [], tables_in_update = [], sql_connection = None):
		self._sql_update = sql_update
		self._sql_values = sql_values
		self._sql_where = sql_where
		self._sql_tables = sql_tables
		self._tables_in_update = tables_in_update 
		self._sql_connection = sql_connection

	@property
	def _total_tables_in_update (self):
		return sum(self._tables_in_update)

	@property
	def _sql_tables_in_update (self):
		return [sql_table for table_bool, sql_table in zip(self._tables_in_update, self._sql_tables) if table_bool]

	def update (self):

		# Update the values
		if self._sql_where:
			if self._total_tables_in_update > 1:
				for parent_key, foreign_key in foreignKeyPairs(self._sql_tables_in_update):
					if not self._sql_where:
						self._sql_where = SQLWhere.fromQuery(self._sql_tables)
					self._sql_where.addEQColWhere(parent_key, foreign_key)
			self._sql_connection.execute(self._sql_update.update().values(**self._sql_values).where(self._sql_where.where_statement))
		else:
			self._sql_connection.execute(self._sql_update.update().values(**self._sql_values))

	@classmethod
	def fromConfig (cls, config_data, sql_connection):

		# Create a SQLQuery object from the config file and sql_connection
		return cls(sql_update = None, 
				   sql_values = {}, 
				   sql_where = [], 
				   sql_tables = config_data.sql_tables, 
				   tables_in_update = [False] * len(config_data.sql_tables), 
				   sql_connection = sql_connection)

	def addTableToUpdate(self, table):

		# Add the table, update the tables in use
		for table_pos, sql_table in enumerate(self._sql_tables):
			if table == sql_table:
				self._tables_in_update[table_pos] = True
		if self._sql_update: raise Exception('Table (%s) already assigned' % self._sql_update)
		self._sql_update = table

	def addDictValues(self, col_dict):

		# Add the update from a dict, update the tables in use
		sql_values = SQLValues.fromColDictValues(self._sql_update, col_dict)
		self._sql_values = sql_values

	'''
	def addDictWhere(self, where_dict, include = None, type = None, table_dict = False):

		# Add the where statement from a dict, update the tables in use
		if not self._sql_where:
			self._sql_where = SQLWhere.fromQuery(self._sql_tables)
		self._sql_where.addEQColDictWhere(where_dict)
		self._sql_where.addTableDictWhere(where_dict)
		for table_pos in self._sql_where._tables_used:
			if not self._tables_in_update[table_pos]:
				self._tables_in_update[table_pos] = True
	'''

	def addDictWhere (self, where_dict, include = None, cmp_type = None, dict_type = None):

		if not include: raise Exception('Must indicate include/exclude type for where statement')
		if not cmp_type: raise Exception('Must indicate comparison type for where statement')
		if not dict_type: raise Exception('Must indicate dict type for where statement')

		# Add the where statement from a dict, update the tables in use
		if not self._sql_where:
			self._sql_where = SQLWhere.fromQuery(self._sql_tables)
		if dict_type == 'Column': self._sql_where.addColDictWhere(where_dict, include = include, cmp_type = cmp_type)
		elif dict_type == 'Table': self._sql_where.addTableDictWhere(where_dict, include = include, cmp_type = cmp_type)
		for table_pos in self._sql_where._tables_used:
			if not self._tables_in_update[table_pos]:
				self._tables_in_update[table_pos] = True

class SQLInsert ():
	def __init__ (self, sql_insert = None, sql_values = [], sql_tables = [], tables_in_insert = [], sql_ignore_constraint = '', sql_update_constraint = '', sql_connection = None, sql_type = ''):
		self._sql_insert = sql_insert
		self._sql_values = sql_values
		self._sql_tables = sql_tables
		self._tables_in_insert = tables_in_insert
		self._sql_ignore_constraint = sql_ignore_constraint
		self._sql_connection = sql_connection
		self._sql_type = sql_type

	def insert (self):

		self._sql_connection.execute(insert(self._sql_insert).values(self._sql_values))

	def insertIgnore (self):

		if self._sql_type == 'postgresql': 
			self._sql_connection.execute(postgresql_insert(self._sql_insert).values(self._sql_values).on_conflict_do_nothing(constraint = self._sql_ignore_constraint))
		else: 
			self._sql_connection.execute(insert(self._sql_insert).values(self._sql_values).prefix_with('OR IGNORE'))

	@classmethod
	def fromConfig (cls, config_data, sql_connection):

		# Create a SQLQuery object from the config file and sql_connection
		return cls(sql_insert = None, 
				   sql_values = [], 
				   sql_tables = config_data.sql_tables, 
				   tables_in_insert = [False] * len(config_data.sql_tables),
				   sql_connection = sql_connection,
				   sql_type = config_data.type)

	def addIgnore (self, col_name):

		try: getattr(self._sql_insert.columns, col_name)
		except: raise Exception(f'Unable to assign column ({col_name})')
		if self._sql_ignore_constraint: raise Exception('Append string already defined')

		# Assign the schema basename (no reference to database)
		schema_basename = str(self._sql_insert)
		schema_basename = schema_basename if '.' not in schema_basename else schema_basename.split('.')[-1]
		self._sql_ignore_constraint = f'{schema_basename}_{col_name}_key'

	def addTableToInsert (self, table):

		# Add the table, update the tables in use
		for table_pos, sql_table in enumerate(self._sql_tables):
			if table == sql_table:
				self._tables_in_insert[table_pos] = True
		if self._sql_insert: raise Exception('Table (%s) already assigned' % self._sql_insert)

		self._sql_insert = table

	def addDictValues (self, col_dict):
		
		# Confirm the data is within a list
		if not isinstance(col_dict, dict): raise Exception('Values must be stored within a dict')

		# Add the insert from a dict
		sql_values = SQLValues.fromColDictValues(self._sql_insert, col_dict)
		self._sql_values.append(sql_values)

	def addListValues (self, col_dict_list):
		
		# Confirm the data is within a list
		if not isinstance(col_dict_list, list): raise Exception('Values must be stored within a List')

		# Add the insert from a list of dict
		for col_dict in col_dict_list:
			self.addDictValues(col_dict)

	def addDataFrameValues (self, dataframe):

		# Convert the dataframe
		col_dict_list = dataframe.to_dict('records')

		# Add the values
		self.addListValues(col_dict_list)
		
class SQLValues (dict):
	def __init__(self, *arg, **kw):
		super(SQLValues, self).__init__(*arg, **kw)

	@classmethod
	def fromColDictValues (cls, sql_table, col_dict):

		# Add the column-based dict values
		for col_name, col_value in col_dict.items():
			try: 
				getattr(sql_table.columns, col_name)
			except: raise Exception('Unable to assign value: %s: %s' % (col_name, col_value))
		# Create a SQLValues object
		return cls(col_dict)

class SQLWhere ():
	def __init__ (self, sql_tables = [], tables_in_where = []):
		self._sql_where = []
		self._tables_in_where = tables_in_where
		self._sql_tables = sql_tables

	@property
	def where_statement(self):
		if len(self._sql_where) == 1: return self._sql_where[0]
		else: return operator.and_(*self._sql_where)

	@property
	def _tables_used (self):
		return [table_pos for table_pos, table_bool in enumerate(self._tables_in_where) if table_bool]

	def __contains__ (self, new_where):

		# Check if the object contains the where statement
		for sql_where in self._sql_where:
			if new_where.compare(sql_where): return True
		else: return False

	def addTableDictWhere (self, table_dict, include, cmp_type):

		table_dict_pos, table_dict_arg = None, None
		for table_name, col_dict in table_dict.items():
			for table_pos, sql_table in enumerate(self._sql_tables):
				if str(sql_table) == table_name: 
					table_dict_arg = sql_table
					table_dict_pos = table_pos
			if table_dict_pos == None: raise Exception('Unable to assign table: %s' % table_name)
			if not self._tables_in_where[table_dict_pos]: self._tables_in_where[table_dict_pos] = True
			if cmp_type == 'EQ': self.addEQTableDictWhere(table_dict_arg, col_dict, include)
			elif cmp_type == 'IN': self.addINTableDictWhere(table_dict_arg, col_dict, include)
			elif cmp_type == 'LIKE': self.addLIKETableDictWhere(table_dict_arg, col_dict, include)

	def addEQTableDictWhere (self, sql_table, col_dict, include):

		# Add the table-based dict where
		for col_name, col_value in col_dict.items():
			where_arg = None
			where_passed = False
			try:
				test_arg = getattr(sql_table.columns, col_name)
				if len(test_arg.foreign_keys) >= 1:
					continue
				where_arg = test_arg
				where_passed = True
			except: pass
			if not where_passed: raise Exception('Unable to assign where: %s: %s' % (col_name, col_value))
			if include: self._sql_where.append(where_arg == col_value)
			else: self._sql_where.append(where_arg != col_value)

	def addLIKETableDictWhere (self, sql_table, col_dict, include):

		# Add the table-based dict where
		for col_name, col_value in col_dict.items():
			where_arg = None
			where_passed = False
			try:
				test_arg = getattr(sql_table.columns, col_name)
				if len(test_arg.foreign_keys) >= 1:
					continue
				where_arg = test_arg
				where_passed = True
			except: pass
			if not where_passed: raise Exception('Unable to assign where: %s: %s' % (col_name, col_value))
			if include: self._sql_where.append(where_arg.like(col_value))
			else: self._sql_where.append(where_arg.notilike(col_value))

	def addINTableDictWhere (self, sql_table, col_dict, include):

		# Add the table-based dict where
		for col_name, col_values in col_dict.items():
			if not isinstance(col_values, list): Exception('Column values are not list')
			where_arg = None
			where_passed = False
			try:
				test_arg = getattr(sql_table.columns, col_name)
				if len(test_arg.foreign_keys) >= 1:
					continue
				where_arg = test_arg
				where_passed = True
			except: pass
			if not where_passed: raise Exception('Unable to assign where: %s: %s' % (col_name, col_values))
			if include: self._sql_where.append(where_arg.in_(col_values))
			else: self._sql_where.append(where_arg.notin_(col_values))

	def addColDictWhere (self, col_dict, include, cmp_type):

		if cmp_type == 'EQ': self.addEQColDictWhere(col_dict, include)
		elif cmp_type == 'IN': self.addINColDictWhere(col_dict, include)
		elif cmp_type == 'LIKE': self.addLIKEColDictWhere(col_dict, include)


	def addEQColDictWhere (self, col_dict, include):

		# Add the column-based dict where
		for col_name, col_value in col_dict.items():
			where_pos, where_arg = None, None
			for table_pos, sql_table in enumerate(self._sql_tables):
				try:
					test_arg = getattr(sql_table.columns, col_name)
					if len(test_arg.foreign_keys) >= 1:
						continue
					where_arg = test_arg
					where_pos = table_pos
				except: pass
			if where_pos == None: raise Exception('Unable to assign where: %s: %s' % (col_name, col_value))
			if not self._tables_in_where[where_pos]: self._tables_in_where[where_pos] = True
			if include: self._sql_where.append(where_arg == col_value)
			else: self._sql_where.append(where_arg != col_value)

	def addLIKEColDictWhere (self, col_dict, include):

		# Add the column-based dict where
		for col_name, col_value in col_dict.items():
			where_pos, where_arg = None, None
			for table_pos, sql_table in enumerate(self._sql_tables):
				try:
					test_arg = getattr(sql_table.columns, col_name)
					if len(test_arg.foreign_keys) >= 1:
						continue
					where_arg = test_arg
					where_pos = table_pos
				except: pass
			if where_pos == None: raise Exception('Unable to assign where: %s: %s' % (col_name, col_value))
			if not self._tables_in_where[where_pos]: self._tables_in_where[where_pos] = True
			if include: self._sql_where.append(where_arg.like(col_value))
			else: self._sql_where.append(where_arg.notilike(col_value))

	def addINColDictWhere (self, col_dict, include):

		# Add the column-based dict where
		for col_name, col_values in col_dict.items():
			if not isinstance(col_values, list): Exception('Column values are not list')
			where_pos, where_arg = None, None
			for table_pos, sql_table in enumerate(self._sql_tables):
				try:
					test_arg = getattr(sql_table.columns, col_name)
					if len(test_arg.foreign_keys) >= 1:
						continue
					where_arg = test_arg
					where_pos = table_pos
				except: pass

			if where_pos == None: raise Exception('Unable to assign where: %s: %s' % (col_name, col_values))
			if not self._tables_in_where[where_pos]: self._tables_in_where[where_pos] = True
			if include: self._sql_where.append(where_arg.in_(col_values))
			else: self._sql_where.append(where_arg.notin_(col_values))
			
	def addEQColWhere (self, col1, col2):

		# Add the column eq were (i.e. col1 == col2)
		new_where = (col1 == col2)
		if new_where not in self._sql_where:
			self._sql_where.append(new_where)
		else:
			logging.warning('Where (%s) already assigned' % str(new_where))

	@classmethod
	def fromConfig (cls, config_data):

		# Create a SQLWhere object
		return cls(sql_tables = config_data.sql_tables, tables_in_where = [False] * len(config_data.sql_tables))

	@classmethod
	def fromQuery (cls, sql_tables):

		# Create a SQLWhere object
		return cls(sql_tables = sql_tables, 
				   tables_in_where = [False] * len(sql_tables))

'''
def currentTime(timezone = 'US/Eastern'):

	# Set the time format for the logging system
	time_format = '%Y-%m-%d %H:%M:%S %Z'

	# Set the timezone
	specified_timezone = pytz.timezone(timezone)

	# Set the current time
	current_time = datetime.datetime.now(specified_timezone)
		
	# Return string formatted time
	return current_time.strftime(time_format)

def valueMarksStr (values):

	# Determine the number of values
	value_count = len(values)

	# Return the question mark string
	return ', '.join(['?'] * value_count)

def quoteStr (str_to_quote):

	# Create list of characters that require quotes
	quote_chars = [' ', '-']

	try:
		# Check if any of the quote characters are within the string
		if any(quote_char in str_to_quote for quote_char in quote_chars):

			# Confirm the string isn't already quoted
			if str_to_quote[0] != '"' and str_to_quote[-1] != '"': 

				# Update the string with quotes
				str_to_quote = '"%s"' % str_to_quote

	except:

		pass
		
	# Return the string
	return str_to_quote

def quoteField (field_str, split_by_dot = True):

	# Check if the field should be split by a dot '.' symbol
	if split_by_dot:

		# Create a string to rebuild the split items
		joined_str = ''

		# Split the string by a dot '.' symbol
		for field_sub_str in field_str.split('.', 1):

			# Check if the join string is empty
			if joined_str:

				# Add a dot '.' symbol
				joined_str += '.'

			# Quote the entire string
			joined_str += quoteStr(field_sub_str)

		# Return the joined string
		return joined_str

	else:

		# Quote the entire string
		field_str = quoteStr(field_str)

		# Return the field string
		return field_str

def quoteFields (field_list, split_by_dot = True):

	# Loop the list by index
	for pos in range(len(field_list)):

		# Quote the current item in the list
		field_list[pos] = quoteField(field_list[pos], split_by_dot = split_by_dot)

	# Return the quoted list
	return field_list

def returnSetExpression (set_dict):

	# Create an empty string to hold the set expression
	set_expression_str = ''

	# Loop the set dict
	for column in set_dict.keys():

		# Check if the set expression is not empty
		if set_expression_str:

			# Add a comma
			set_expression_str += ', '

		set_expression_str += '%s = ?' % quoteField(column)

	# Return the set expression
	return set_expression_str

def returnSelectionDict (selection_dict):

	# Create the selection string
	selection_str = ''

	# Loop the dict by operator
	for selection_operator, selection_data in selection_dict.items():

		# Loop the selection dict
		for column, value_list in selection_data.items():

			# Check that the selection string isn't empty
			if selection_str:

				# Add the AND operator between each statement
				selection_str += ' AND '

			# Check if the selection operator is either IN or NOT IN
			if 'IN' in selection_operator:
					
				# Add the quoted column, the selection operator, and the question marks to the string
				selection_str += '%s %s (%s)' % (quoteField(column), selection_operator, valueMarksStr(value_list))

			# Check if the selection operator is either LIKE or NOT LIKE
			elif 'LIKE' in selection_operator:

				# Create a selection substring
				selection_sub_str = ''

				# Loop the selecion data
				for value_item in value_list:

					# Check that the selection string isn't empty
					if selection_sub_str:

						# Add the AND operator between each statement
						selection_sub_str += ' OR '

					# Add the quoted column and the selection operator
					selection_sub_str += '%s %s ?' % (quoteField(column), selection_operator)

				# Check if the sub string should be within parentheses
				if len(value_list) > 1:

					# Enclose the LIKE statements
					selection_sub_str = '(%s)' % selection_sub_str

				# Add the substring to the string
				selection_str += selection_sub_str

	# Update log
	logging.info('Created selection statement for database call')

	return selection_str

def returnSelectionValues (selection_dict):

	# Create an empty list to hold the values for the selection expression
	expression_statement_values = []

	# Loop the dict by operator
	for selection_operator, selection_data in selection_dict.items():

		# Check if the selection operator is either IN or NOT IN
		if 'IN' in selection_operator:

			# Add the include expression values to the list
			expression_statement_values.extend(itertools.chain.from_iterable(list(selection_data.values())))

		# Check if the selection operator is either LIKE or NOT LIKE
		elif 'LIKE' in selection_operator:

			# Loop each value in the selection dict
			for selection_value in itertools.chain.from_iterable(list(selection_data.values())):

				# Add the include expression value with wildcards to the list
				expression_statement_values.append('%' + selection_value + '%')

	# Quote the values
	expression_statement_values = quoteFields(expression_statement_values)

	# Update log
	logging.info('Successfully created value list for selection statement')

	# Return the list
	return expression_statement_values

def innerJoinTables (table_list, join_column_list):

	# Add the first table to the string
	inner_join_str = table_list[0]

	# Loop the tables, skipping the first entry
	for table_name, table_join_column in zip(table_list[1:], join_column_list):

		# Confirm the data is a linking table
		if isinstance(table_name, dict):

			# Assign the linking table and join column
			for link_table_name, link_table_join_column in zip(table_name.keys(), table_join_column.keys()):

				# Assign the sub table data
				sub_table_list = table_name[link_table_name]
				sub_join_column_list = table_join_column[link_table_join_column]

				# Add the first table to the string
				sub_inner_join_str = sub_table_list[0]

				# Loop the tables, skipping the first entry
				for sub_table_name, sub_table_join_column in zip(sub_table_list[1:], sub_join_column_list):

					# Add the sub inner join string
					sub_inner_join_str += ' INNER JOIN {0} ON {1}.{2} = {0}.{2}'.format(sub_table_name, sub_table_list[0], quoteField(sub_table_join_column))

				# Add the inner join string
				inner_join_str +=  ' INNER JOIN ({0}) {1} ON {2}.{3} = {1}.{3}'.format(sub_inner_join_str, link_table_name, table_list[0], quoteField(link_table_join_column))
		else:

			# Add the inner join string
			inner_join_str += ' INNER JOIN {0} ON {1}.{2} = {0}.{2}'.format(table_name, table_list[0], quoteField(table_join_column))

	# Update log
	logging.info('Created join tables statement for database call')

	# Return the inner join string
	return inner_join_str

def reportSqlError (sql_error, column_list = None, value_list = None):

	# Check if the unique constraint failed
	if 'UNIQUE constraint failed' in str(sql_error):

		# Get the column path that failed
		sql_column_path = str(sql_error).split(': ')[1]

		# Assign the table
		sql_table = sql_column_path.split('.')[0]

		# Assign the column
		sql_column = sql_column_path.split('.')[1]

		# Create an empty warning string to add to
		sql_error_message = ''

		# Check if column and values were given, and that the sql column can be found
		if column_list and value_list and (sql_column_path in column_list or sql_column in column_list):

			# Check if the column path is within the value list
			if sql_column_path in column_list:

				# Get the column index
				sql_column_index = column_list.index(sql_column_path)

			# Check if the column is within the value list
			elif sql_column in column_list:

				# Get the column index
				sql_column_index = column_list.index(sql_column)

			# Update the message
			sql_error_message += '%s already exists' % value_list[sql_column_index]
		
		# Check if the error message has been populated
		if sql_error_message:

			# Add punctuation
			sql_error_message += '. '

		# Quote the column path, if needed
		sql_quoted_column_path = quoteField(sql_column_path)

		# Display warning for duplicates, as raise would only report the first entry
		logging.warning('%s%s does not support duplicate entries' % (sql_error_message, sql_quoted_column_path))

	# If not a known error, report the standard message
	else:

		raise sql_error

def createTable (cursor, table, column_assignment_str):

	try:
		
		# Createt the table assignment string
		sqlite_create_table = 'CREATE TABLE IF NOT EXISTS %s (%s)' % (table, column_assignment_str)

		# Execute the create table command
		cursor.execute(sqlite_create_table)

		# Update log
		logging.info('Successfully created table (%s) in database' % table)

	except sqlite3.Error as error:
		raise Exception(error)

def insertValues (cursor, table, column_list, value_list):

	try:

		# Create list with column_list and the date columns
		column_list_w_dates = column_list + ['Last Modified (%s)' % table, 'Entry Created (%s)' % table]

		# Quote the columns as needed
		column_list_w_dates = quoteFields(column_list_w_dates)

		# Record the current time
		insert_time = currentTime()

		# Create list with value_list and the date values
		value_list_w_dates = value_list + [insert_time, insert_time]

		# Quote the values as needed
		value_list_w_dates = quoteFields(value_list_w_dates, split_by_dot = False)
	
		# Create the insert string
		sqlite_insert_values = 'INSERT INTO %s (%s) VALUES (%s)' % (table, ', '.join(column_list_w_dates), valueMarksStr(column_list_w_dates))

		# Execute the insert values command
		cursor.execute(sqlite_insert_values, value_list_w_dates)

		# Update log
		logging.info('Successfully inserted data into the database')

	except sqlite3.Error as sql_error:

		# Report the error
		reportSqlError(sql_error, column_list = column_list, value_list = value_list)

def updateValues (cursor, table, selection_dict, update_dict, update_table_column = None, tables_to_join = None, join_table_columns = None):

	try:

		# Record the current time
		insert_time = currentTime()

		# Create the select expression string for the SQL call
		select_expression = returnSelectionDict(selection_dict)

		# Create a list of the values associated with the select expression string
		select_value_list = returnSelectionValues(selection_dict)

		# Create an empty from expression for single table updates
		from_expression = ''

		# Check for multiple tables to select from
		if update_table_column and tables_to_join and join_table_columns:

			# Create an inner join statement
			table_str = innerJoinTables(tables_to_join, join_table_columns)

			# Update the selection expression with a select clause
			select_expression = '{1} IN (SELECT {0}.{1} FROM {2} WHERE {3})'.format(table, quoteField(update_table_column), table_str, select_expression)

		# Check if the multiple tables were incorrectly assigned
		elif update_table_column or tables_to_join or join_table_columns:

			raise Exception('Error updating database, unable to create FROM statement')

		# Copy the update dict
		update_dict_w_date = copy.copy(update_dict)

		# Update the update dict to include the date modified column
		update_dict_w_date['Last Modified ({0})'.format(table)] = insert_time

		# Create the update expression string for the SQL call
		update_expression = returnSetExpression(update_dict_w_date)

		# Create a quoated list of the values associated with the update expression string
		update_value_list = quoteFields(list(update_dict_w_date.values()), split_by_dot = False)

		# Combine the value lists and save as a tuple
		sqlite_value_list = tuple(update_value_list + select_value_list)

		# Create the update string
		sqlite_update_values = 'UPDATE %s SET %s WHERE %s' % (table, update_expression, select_expression)

		# Execute the insert values command
		cursor.execute(sqlite_update_values, sqlite_value_list)

		# Update log
		logging.info('Successfully updated the database')

	except sqlite3.Error as error:

		raise Exception(error)

def retrieveValues (cursor, tables, selection_dict, column_list, join_table_columns = None):

	try:

		# Create the select expression string for the SQL call
		select_expression = returnSelectionDict(selection_dict)

		# Create a list of the values associated with the select expression string
		select_value_list = returnSelectionValues(selection_dict)

		# Create a quoated list of the values
		select_value_list = quoteFields(select_value_list, split_by_dot = False)

		# Quote the columns as needed
		column_list = quoteFields(column_list)

		#  a string of the column list
		column_str = ', '.join(column_list)

		# Check if there is more than a single table
		if len(tables) > 1:

			# Create an inner join statement
			table_str = innerJoinTables(tables, join_table_columns)

		else:

			# Define the table string
			table_str = tables[0]
	
		# Create the basic select string
		sqlite_select_values = 'SELECT %s FROM %s' % (column_str, table_str)

		# Check if a selection expression was given
		if select_expression:

			# Update the sqlite string if an expression was given
			sqlite_select_values += ' WHERE %s' % select_expression

		# Check if selection expression values were created
		if select_value_list:

			# Execute the select values command
			cursor.execute(sqlite_select_values, select_value_list)

		else:

			# Execute the select values command
			cursor.execute(sqlite_select_values)

		# Save the selected results
		selection_results = cursor.fetchall()

		# Check if anything was returned
		if len(selection_results) == 0:

			# Print error message
			raise Exception('No output produced. Please check your input')

		# Update log
		logging.info('Successfully retrieved results from database')

		# Return the retrieved data as a dict
		return selection_results

	except sqlite3.Error as error:
		raise Exception(error)

def confirmValue (cursor, table, column, value):
	
	try:

		# Create a list to store the check results
		check_results = []

		# Check if the value is None
		if value == None:
			
			# Create the value check string
			sqlite_check_value = 'SELECT COUNT(*) FROM %s WHERE %s is NULL' % (table, quoteField(column))

			# Execute the create table command
			cursor.execute(sqlite_check_value)

		else:
		
			# Create the value check string
			sqlite_check_value = 'SELECT COUNT(*) FROM %s WHERE %s = ?' % (table, quoteField(column))

			# Execute the create table command
			cursor.execute(sqlite_check_value, [quoteField(value)])

		# Save the selected results
		selection_results = cursor.fetchall()

		# Check if there was an assignment error
		if not selection_results:
			raise Exception('Assignment error when checking for presence of %s in %s:%s)' % (value, table, column))

		# Check if the table was found
		if selection_results[0][0] >= 1:

			# Return True
			return True

		# Check if the table was not found
		elif selection_results[0][0] == 0:

			# Return False
			return False

		# Check if there was an unknown error
		else:

			raise Exception('Unknown error checking for presence of %s in %s:%s)' % (value, table, column))

	except sqlite3.Error as error:
		raise Exception(error)

def backupDatabase (database, backup_database):

	# Connect to the databse 
	sqlite_database_connection = sqlite3.connect(database)

	# Connect to the backup
	sqlite_backup_connection = sqlite3.connect(backup_database)

	# Backup the database
	with sqlite_backup_connection:
		sqlite_database_connection.backup(sqlite_backup_connection, pages=1)
	
	# Close the connections 	
	sqlite_backup_connection.close()
	sqlite_database_connection.close()
'''

'''
class SQLQuery ():
	def __init__ (self, sql_query = [], sql_filters = None, sql_updates = {}, sql_tables = [], tables_in_query = [], sql_session = None):
		self.query_results = None
		self._sql_query = sql_query
		self._sql_filters = sql_filters
		self._sql_updates = sql_updates
		self._sql_tables = sql_tables
		self._tables_in_query = tables_in_query
		self._sql_session = sql_session

	@property
	def _total_tables_in_query (self):
		return sum(self._tables_in_query)

	@property
	def _total_tables_in_results (self):
		return len(self._sql_query)

	def query_all (self):

		if self._total_tables_in_query > 1:
			for parent_key, foreign_key in foreignKeyPairs(self._sql_tables):
				if not self._sql_filters:
					self._sql_filters = SQLFilter.fromQuery(self._sql_tables)
				self._sql_filters.addColEQFilter(parent_key, foreign_key)
		self.query_results = self._sql_session.query(*self._sql_query).filter(*self._sql_filters.filters).all()

	def query_update (self):

		if self._total_tables_in_query > 1:
			for parent_key, foreign_key in foreignKeyPairs(self._sql_tables):
				if not self._sql_filters:
					self._sql_filters = SQLFilter.fromQuery(self._sql_tables)
				self._sql_filters.addColEQFilter(parent_key, foreign_key)
		self.query_results = self._sql_session.query(*self._sql_query).filter(*self._sql_filters.filters).all()

		if self._total_tables_in_results > 1:

			for row in self.query_results:
				#session.add(prop)
				#row_dict = {}
				for result_table in row:
					for update_table in self._sql_updates.updates:
						if result_table.__tablename__ != update_table.__tablename__:
							continue
							for key, value in self._sql_updates.updates[update_table].items():
								key = value
								self._sql_session.add(key)


								#print (key == result_table)


					#session.add(prop)
					#for result_key, result_value in self.object_as_dict(result_table).items():
					#	row_dict[result_key] = result_value
				#print(row_dict)
		else:

			for row in self.query_results:
				print('Here')
				#print(self.object_as_dict(row))

	def toCSV (self):

		if self._total_tables_in_results > 1:

			for row in self.query_results:
				row_dict = {}
				for result_table in row:
					for result_key, result_value in self.object_as_dict(result_table).items():
						row_dict[result_key] = result_value
				print(row_dict)
		else:

			for row in self.query_results:
				print(self.object_as_dict(row))

	@classmethod
	def fromConfig (cls, config_data, sql_session):

		# Create a SQLQuery object from the config file and sql_session
		return cls(sql_tables = config_data.sql_tables, tables_in_query = [False] * len(config_data.sql_tables), sql_session = sql_session)

	def addQuery(self, query, type = None):

		# Add the query, update the tables in use
		for table_pos, sql_table in enumerate(self._sql_tables):
			if query == sql_table:
				self._tables_in_query[table_pos] = True
		self._sql_query.append(query)

	def addDictFilter(self, col_dict, include = None, type = None, table_dict = False):

		# Add the filter from a dict, update the tables in use
		sql_filter = SQLFilter.fromQuery(self._sql_tables)
		sql_filter.addColDictFilter(col_dict)
		for table_pos in sql_filter._tables_used:
			if not self._tables_in_query[table_pos]:
				self._tables_in_query[table_pos] = True
		self._sql_filters = sql_filter

	def addDictUpdate(self, col_dict):

		# Add the update from a dict, update the tables in use
		sql_update = SQLUpdate.fromQuery(self._sql_tables)
		sql_update.addColDictUpdate(col_dict)
		for table_pos in sql_update._tables_used:
			if not self._tables_in_query[table_pos]:
				self._tables_in_query[table_pos] = True
		self._sql_updates = sql_update

	@staticmethod
	def object_as_dict(obj):
		return {c.key: getattr(obj, c.key) for c in inspect(obj).mapper.column_attrs}

class SQLFilter ():
	def __init__ (self, sql_tables = [], tables_in_filters = []):
		self._filters = []
		self._tables_in_filters = tables_in_filters
		self._sql_tables = sql_tables

	@property
	def filters (self):
		return self._filters

	@property
	def _tables_used (self):
		return [table_pos for table_pos, table_bool in enumerate(self._tables_in_filters) if table_bool]

	def __contains__ (self, new_filter):

		# Check if the object contains the filter
		for sql_filter in self._filters:
			if new_filter.compare(sql_filter): return True
		else: return False

	def addColDictFilter (self, col_dict):

		# Add the column-based dict filter
		for col_name, col_value in col_dict.items():
			filter_pos, filter_arg = None, None
			for table_pos, sql_table in enumerate(self._sql_tables):
				try:
					test_arg = getattr(sql_table, col_name)
					if len(test_arg.foreign_keys) >= 1:
						continue
					filter_arg = test_arg
					filter_pos = table_pos
				except: pass
			if not filter_arg: raise Exception('Unable to assign filter: %s: %s' % (col_name, col_value))
			if not self._tables_in_filters[filter_pos]: self._tables_in_filters[table_pos] = True
			self._filters.append(filter_arg == col_value)

	def addColEQFilter (self, col1, col2):

		# Add the column eq filter (i.e. col1 == col2)
		new_filter = (col1 == col2)
		if new_filter not in self:
			self._filters.append(new_filter)
		else:
			logging.warning('Filter (%s) already assigned' % str(new_filter))

	@classmethod
	def fromConfig (cls, config_data):

		# Create a SQLFilter object
		return cls(sql_tables = config_data.sql_tables, tables_in_filters = [False] * len(config_data.sql_tables))

	@classmethod
	def fromQuery (cls, sql_tables):

		# Create a SQLFilter object
		return cls(sql_tables = sql_tables, tables_in_filters = [False] * len(sql_tables))

class SQLUpdate ():
	def __init__ (self, sql_tables = [], tables_in_updates = []):
		self._updates = defaultdict(lambda: defaultdict(str))
		self._tables_in_updates = tables_in_updates
		self._sql_tables = sql_tables

	@property
	def updates (self):
		return self._updates

	@property
	def _tables_used (self):
		return [table_pos for table_pos, table_bool in enumerate(self._tables_in_updates) if table_bool]

	def __contains__ (self, new_update):

		# Check if the object contains the filter
		for update_column in list(new_update):
			if update_column in list(self._updates): return True
		else: return False

	def addColDictUpdate (self, col_dict):

		# Add the column-based dict update
		for col_name, col_value in col_dict.items():
			update_pos, update_arg, update_table = None, None, None
			for table_pos, sql_table in enumerate(self._sql_tables):
				try: 
					test_arg = getattr(sql_table, col_name)
					if len(test_arg.foreign_keys) >= 1:
						continue
					update_arg = test_arg
					update_pos = table_pos
					update_table = sql_table
				except: pass
			if not update_arg: raise Exception('Unable to assign update: %s: %s' % (col_name, col_value))

			if not self._tables_in_updates[update_pos]: self._tables_in_updates[update_pos] = True
			self._updates[update_table][update_arg] = col_value

	@classmethod
	def fromConfig (cls, config_data):

		# Create a SQLFilter object
		return cls(sql_tables = config_data.sql_tables, tables_in_updates = [False] * len(config_data.sql_tables))

	@classmethod
	def fromQuery (cls, sql_tables):

		# Create a SQLFilter object
		return cls(sql_tables = sql_tables, tables_in_updates = [False] * len(sql_tables))
'''


import os
import logging

def confirmExecutable (executable):

	'''
	    Confirm if an executable exists.

	    Parameters
	    ----------
	    executable : str
	        Executable to confirm
	'''

	# Loop potental locations of executables
	for path in os.environ['PATH'].split(os.pathsep):

		# Join current path and executable
		executable_file = os.path.join(path, executable)

		# Check if the executable path exists, and if so, is an executable
		if os.path.isfile(executable_file) and os.access(executable_file, os.X_OK):

			logging.info('Calling executable: %s' % executable_file)

			# Return the path if the executable was found
			return executable_file

	# Return None if the executable was not found
	return None


import os
import logging
import string
import shutil
import gzip

from collections import OrderedDict
from Bio import SeqIO

from kocher_tools.fastq_multx import i5BarcodeJob, i7BarcodeJob
from kocher_tools.vsearch import *
from kocher_tools.compress import gzipCompress, gzipIsEmpty

class Multiplex (list):
	def __init__ (self, *arg, **kw):
		super(Multiplex, self).__init__(*arg, **kw)
		self.i5_file = ''
		self.i7_file = ''
		self.R1_file = ''
		self.R2_file = ''
		self.out_path = ''
		self.discard_empty_output = True

	def __contains__ (self, plate_str):
		if plate_str in [str(plate) for plate in self]:
			return True
		else:
			return False

	def __getitem__(self, plate_str):
		for plate in self:
			if str(plate) == plate_str:
				return plate

		raise Exception('%s not found in Multiplex' % plate_str) 

	@property
	def files (self):
		# Return files
		return [file for file in [self.i5_file, self.i7_file, self.R1_file, self.R2_file]]

	def assignOutputPath (self, out_path):

		# Assign the output path
		self.out_path = out_path

		# Create the directory, if needed
		if not os.path.exists(self.out_path):
			os.makedirs(self.out_path)

		logging.info('Output directory assigned: %s' % out_path)

	def assignFiles (self, i5_barcode_file, i7_barcode_file, r1_file, r2_file):
		self.i5_file = i5_barcode_file
		self.i7_file = i7_barcode_file
		self.R1_file = r1_file
		self.R2_file = r2_file

	def assignPlates (self, i5_map_filename):

		# Check if files have been assigned 
		if not self.files:
			
			# Raise an excpetion
			raise Exception('File assignment required for plate assignment')

		# Open the i5 map file
		with open(i5_map_filename) as i5_map_file:

			# Loop the i5 map file, line by line
			for i5_map_line in i5_map_file:

				# Split the line into: plate, barcode, locus
				i5_plate, i5_barcode, i5_locus = i5_map_line.split()

				# Assign the plate
				plate_object = Plate()

				# Assign the discard status
				plate_object.discard_empty_output = self.discard_empty_output

				# Assign the plate output path
				plate_object.out_path = self.out_path

				# Assign the plate name
				plate_object.name = i5_plate

				# Assign the locus
				plate_object.locus = i5_locus

				# Assign the filenames
				plate_object.assignFilenames()

				# Append the plate
				self.append(plate_object)

	def movePlates (self):

		# Loop the plates
		for plate in self:

			# Move the files
			plate.moveFiles()

	def deMultiplex (self, i5_map_filename):

		# Use the i5 map to demultiplex
		i5BarcodeJob(i5_map_filename, self.i5_file, self.i7_file, self.R1_file, self.R2_file, self.out_path, self.discard_empty_output)

	def compileMostAbundant (self, out_filename, out_format = 'fasta'):
		
		# Open the output file
		compiled_file = open(out_filename, 'w')

		# Loop each plate
		for plate in self:

			# Get the common fasta records from the plate
			for fasta_record in plate.yieldMostAbundant():

				# Write the 
				compiled_file.write(fasta_record.format(out_format)) 

		# Close the file
		compiled_file.close()

	def removeUnmatched (self):

		# Define an empty output path as default
		out_path = ''

		# Check if an output path has been defined
		if self.out_path:

			# Add a trailing directory symbol
			out_path = os.path.join(self.out_path, '')

		# Check if the empty output files were created
		if not self.discard_empty_output:

			# Remove the unmatched i5 output
			os.remove('%sunmatched_%s.fastq.gz' % (out_path, 'i5')) 

		# Remove the unmatched i7, R1, and R2 output
		os.remove('%sunmatched_%s.fastq.gz' % (out_path, 'i7')) 
		os.remove('%sunmatched_%s.fastq.gz' % (out_path, 'R1')) 
		os.remove('%sunmatched_%s.fastq.gz' % (out_path, 'R2'))

class Plate (list):
	def __init__ (self, *arg, **kw):
		super(Plate, self).__init__(*arg, **kw)
		self.name = ''
		self.locus = ''
		self.plate_i5_file = ''
		self.plate_i7_file = ''
		self.plate_R1_file = ''
		self.plate_R2_file = ''
		self.out_path = ''
		self.discard_empty_output = None

	def __str__(self):

		# Return the name of the plate
		return self.name

	def __contains__ (self, well_str):
		if well_str in [str(well) for well in self]:
			return True
		else:
			return False

	def __getitem__(self, well_str):
		for well in self:
			if str(well) == well_str:
				return well

		raise Exception('%s not found' % well_str) 

	@property
	def files (self):

		# Return the files, if assigned
		return [file for file in [self.plate_i5_file, self.plate_i7_file, self.plate_R1_file, self.plate_R2_file] if file]

	@property
	def well_path(self):

		# Assign the well path, using the out_path, name, and locus
		return os.path.join(self.out_path, self.name, self.locus)
	
	def assignFilenames (self):

		# Define an empty output path as default
		plate_out_path = ''

		# Check if an output path has been defined
		if self.out_path:

			# Add a trailing directory symbol
			plate_out_path = os.path.join(self.out_path, '')

		# Check if the empty output should be assigned
		if self.discard_empty_output == False:

			# Assign the i5 filename, by inserting the output path, plate name, and locus
			self.plate_i5_file = '%s%s_%s_i5.fastq.gz' % (plate_out_path, self.name, self.locus)

		# Assign the other filenames, by inserting the output path, plate name, and locus
		self.plate_i7_file = '%s%s_%s_i7.fastq.gz' % (plate_out_path, self.name, self.locus)
		self.plate_R1_file = '%s%s_%s_R1.fastq.gz' % (plate_out_path, self.name, self.locus)
		self.plate_R2_file = '%s%s_%s_R2.fastq.gz' % (plate_out_path, self.name, self.locus)

	def moveFiles (self):

		# Create the output path
		plate_out_path = os.path.join(self.out_path, self.name, self.locus)

		# Create the output directories
		if not os.path.exists(plate_out_path):
			os.makedirs(plate_out_path)

		# Check if the empty output should be moved
		if self.discard_empty_output == False:

			# Move the i5 file
			self.plate_i5_file = moveFile(self.plate_i5_file, plate_out_path)

		# Move/Rename the files
		self.plate_i7_file = moveFile(self.plate_i7_file, plate_out_path)
		self.plate_R1_file = moveFile(self.plate_R1_file, plate_out_path)
		self.plate_R2_file = moveFile(self.plate_R2_file, plate_out_path)

	def assignWells (self):

		# Check if files have been assigned 
		if not self.files:
			
			# Raise an excpetion
			raise Exception('File assignment required for well assignment')

		# Loop well letters, A to H
		for char in string.ascii_uppercase[:8]:

			# Loop well position, 1 to 12
			for pos in range(1, 13):

				# Save the well ID
				well_ID = char + str(pos)

				# Assign the Well
				well_object = Well()

				# Assing the discard status
				well_object.discard_empty_output = self.discard_empty_output

				# Assign the output path
				well_object.out_path = self.well_path

				# Save the ID
				well_object.ID = well_ID

				# Assign the plate the well is on
				well_object.on_plate = self.name

				# Assign the filenames
				well_object.assignFilenames()

				# Append the well
				self.append(well_object)

	def moveWells(self):

		# Loop the wells
		for well in self:

			# Move the files
			well.moveFiles()

	def deMultiplexPlate (self, i7_map_filename):

		# Use the i7 map to demultiplex
		i7BarcodeJob(i7_map_filename, self.plate_i7_file, self.plate_R1_file, self.plate_R2_file, self.well_path, self.discard_empty_output)

	def yieldMostAbundant (self):

		# Loop each well
		for well in self:

			# Get the common fasta records from the well
			for fasta_record in well.yieldMostAbundant():

				# Yield the the records from the well
				yield fasta_record

	def removeUnmatchedPlate (self):

		# Create the output path
		plate_out_path = os.path.join(self.out_path, self.name, self.locus, '')

		# Check if the empty output files were created
		if not self.discard_empty_output:

			# Remove the unmatched i5 output
			os.remove('%sunmatched_%s.fastq.gz' % (plate_out_path, 'i7')) 

		# Remove the unmatched R1 and R2 output
		os.remove('%sunmatched_%s.fastq.gz' % (plate_out_path, 'R1')) 
		os.remove('%sunmatched_%s.fastq.gz' % (plate_out_path, 'R2'))

class Well ():
	def __init__ (self):

		# General Args
		self.on_plate = ''
		self.ID = ''
		self.well_i7_file = ''
		self.well_R1_file = ''
		self.well_R2_file = ''
		self.out_path = ''
		self.well_dir = 'Demultiplexed'
		self.discard_empty_output = None

		# Merged Args
		self.merged_file = ''
		self.unmerged_R1_file = ''
		self.unmerged_R2_file = ''
		self.merged_dir = 'Merged'

		# Truncated Args
		self.truncated_file = ''
		self.truncated_dir = 'Truncated'

		# Filtered Args
		self.filtered_file = ''
		self.filtered_dir = 'Filtered'

		# Dereplicated Args
		self.dereplicated_file = ''
		self.dereplicated_dir = 'Dereplicated'

		# Clustered Args
		self.clustered_file = ''
		self.clustered_dir = 'Clustered'

		# Clustered Args
		self.common_file = ''
		self.common_dir = 'Common'

	def __str__(self):

		# Return the ID, if str() is used
		return self.ID

	@property
	def files (self):

		# Return the files, if they were defined
		return [file for file in [self.well_i7_file, self.well_R1_file, self.well_R2_file] if file]
	
	def assignFilenames (self):

		# Define an empty output path as default
		well_out_path = ''

		# Check if an output path has been defined
		if self.out_path:

			# Add a trailing directory symbol
			well_out_path = os.path.join(self.out_path, '')

		# Check if the empty output should be created
		if self.discard_empty_output == False:

			# Assign the i7 file
			self.well_i7_file = '%s%s_i7.fastq.gz' % (well_out_path, self.ID)

		# Assign the R1 and R2 filenames
		self.well_R1_file = '%s%s_R1.fastq.gz' % (well_out_path, self.ID)
		self.well_R2_file = '%s%s_R2.fastq.gz' % (well_out_path, self.ID)

	def moveFiles (self):

		# Create the output path
		well_out_path = os.path.join(self.out_path, self.well_dir)

		# Create the output directories
		if not os.path.exists(well_out_path):
			os.makedirs(well_out_path)

		# Check if the empty output should be moved
		if self.discard_empty_output == False:

			# Move the i7 file
			self.well_i7_file = moveFile(self.well_i7_file, well_out_path)

		# Move/Rename the R1 and R2 files
		self.well_R1_file = moveFile(self.well_R1_file, well_out_path)
		self.well_R2_file = moveFile(self.well_R2_file, well_out_path)

	def mergeWell (self):

		# Check if the R1 or R2 files are empty
		if gzipIsEmpty(self.well_R1_file) or gzipIsEmpty(self.well_R2_file):

			self.merged_file = ''
			self.unmerged_R1_file = ''
			self.unmerged_R2_file = ''

		else:

			# Add a trailing directory symbol to the output path
			merged_path = os.path.join(self.out_path, self.merged_dir, '')

			# Create the directory, if needed
			if not os.path.exists(merged_path):
				os.makedirs(merged_path)

			# Define the merged and unmerged files 
			self.merged_file = '%s%s_merged.fastq' % (merged_path, self.ID)
			self.unmerged_R1_file = '%s%s_notmerged_R1.fastq' % (merged_path, self.ID)
			self.unmerged_R2_file = '%s%s_notmerged_R2.fastq' % (merged_path, self.ID)

			# Merge the well R1 and R2 files
			mergePairs(self.ID, self.well_R1_file, self.well_R2_file, self.merged_file, self.unmerged_R1_file, self.unmerged_R2_file)

			# Gzip the files, return the updated filenames
			self.merged_file = gzipCompress(self.merged_file, return_filename = True)
			self.unmerged_R1_file = gzipCompress(self.unmerged_R1_file, return_filename = True)
			self.unmerged_R2_file = gzipCompress(self.unmerged_R2_file, return_filename = True)

	def truncateWell (self):

		# Check if the merged file is empty
		if self.merged_file == '' or gzipIsEmpty(self.merged_file):

			self.truncated_file = ''

		else:

			# Define the truncated path
			truncated_path = os.path.join(self.out_path, self.truncated_dir)

			# Create the directory, if needed
			if not os.path.exists(truncated_path):
				os.makedirs(truncated_path)

			# Define the truncated file
			self.truncated_file = os.path.join(truncated_path, '%s_stripped.fastq' % self.ID)

			# Truncate the merged file
			truncateFastq(self.merged_file, self.truncated_file)

			# Gzip the file, return the updated filename
			self.truncated_file = gzipCompress(self.truncated_file, return_filename = True)

	def filterWell (self):

		# Check if the truncated file is empty
		if self.truncated_file == '' or gzipIsEmpty(self.truncated_file):

			self.filtered_file = ''

		else:

			# Define the filtered path
			filtered_path = os.path.join(self.out_path, self.filtered_dir)

			# Create the directory, if needed
			if not os.path.exists(filtered_path):
				os.makedirs(filtered_path)

			# Define the filtered file
			self.filtered_file = os.path.join(filtered_path, '%s_filtered.fasta' % self.ID)

			# Filter the truncated file
			filterFastq(self.truncated_file, self.filtered_file)

			# Gzip the file, return the updated filename
			self.filtered_file = gzipCompress(self.filtered_file, return_filename = True)

	def dereplicateWell (self):

		# Check if the filtered file is empty
		if self.filtered_file == '' or gzipIsEmpty(self.filtered_file):

			self.dereplicated_file = ''

		else:

			# Define the dereplicated path
			dereplicated_path = os.path.join(self.out_path, self.dereplicated_dir)

			# Create the directory, if needed
			if not os.path.exists(dereplicated_path):
				os.makedirs(dereplicated_path)

			# Define the dereplicated file
			self.dereplicated_file = os.path.join(dereplicated_path, '%s_dereplicated.fasta' % self.ID)

			# Dereplicate the file
			dereplicateFasta(self.on_plate, self.ID, self.filtered_file, self.dereplicated_file)

			# Gzip the file, return the updated filename
			self.dereplicated_file = gzipCompress(self.dereplicated_file, return_filename = True)

	def clusterWell (self):

		# Check if the dereplicated file is empty
		if self.dereplicated_file == '' or gzipIsEmpty(self.dereplicated_file):

			self.clustered_file = ''

		else:

			# Define the clustered path
			clustered_path = os.path.join(self.out_path, self.clustered_dir)

			# Create the directory, if needed
			if not os.path.exists(clustered_path):
				os.makedirs(clustered_path)

			# Define the clustered file
			self.clustered_file = os.path.join(clustered_path, '%s_clustered.fasta' % self.ID) 

			# Cluster the file
			clusterFasta(self.on_plate, self.ID, self.dereplicated_file, self.clustered_file)

			# Gzip the file, return the updated filename
			self.clustered_file = gzipCompress(self.clustered_file, return_filename = True)

	def sortWell (self):

		# Define the sorted path
		sorted_path = os.path.join(self.out_path, self.clustered_dir)

		# Define the clustered file
		sort_file = os.path.join(sorted_path, '%s_sorted_clustered.fasta' % self.ID) 

		# Cluster the file
		sortFasta(self.on_plate, self.ID, self.clustered_file, sort_file)

		# Gzip the file, return the updated filename
		sort_file = gzipCompress(sort_file, return_filename = True)

		logging.info('%s-%s: Sorted clustered file' % (self.on_plate, self.ID))

		# Rename the clustered file
		self.clustered_file = sort_file

	def mostAbundantWell (self):

		# Check if the clustered file is empty
		if self.clustered_file == '' or gzipIsEmpty(self.clustered_file):

			self.common_file = ''

		else:

			# Define the common path
			common_path = os.path.join(self.out_path, self.common_dir)

			# Create the directory, if needed
			if not os.path.exists(common_path):
				os.makedirs(common_path)

			# Define the common file
			self.common_file = os.path.join(common_path, '%s_common.fasta.gz' % self.ID) 

			# Open file to store most common reads 
			common_file = gzip.open(self.common_file, 'wt')

			# Define an int to store the abundance of the read
			most_abundant_count = 0

			# Define an int to store the rank of the most abundance read
			most_abundant_rank = 0

			# Decompress the gzip file
			with gzip.open(self.clustered_file, "rt") as clustered_handle:

				# Loop the clustered file, line by line
				for clustered_line in clustered_handle:

					# Check if the line is a header
					if clustered_line.startswith('>'):

						# Get the abundance
						read_abundance = int(clustered_line.strip().split('=')[1])

						# Check if the abundance is higher than the stored value
						if read_abundance > most_abundant_count:

							# Update the count, and record is more abundant
							most_abundant_count = read_abundance

							# Get the rank
							read_rank = int(clustered_line.strip().split(';')[0].rsplit('_', 1)[1])

							# Update the rank
							most_abundant_rank = read_rank
			
			# Check if the read rank is not correctly ordered
			if most_abundant_rank != 1:

				# Sort the well to correct the order
				self.sortWell()

			# Decompress the gzip file
			with gzip.open(self.clustered_file, "rt") as clustered_handle:

				# Loop the clustered file, record by record
				for record in SeqIO.parse(clustered_handle, "fasta"):

					# Assign the abundance
					read_abundance = int(record.id.split('=')[1])

					# Check if the abundance is higher than the stored value
					if read_abundance >= (0.5 * most_abundant_count):

						# Write the fasta sequence to the common file
						common_file.write(record.format("fasta"))

			# Close the common file
			common_file.close()

	def yieldMostAbundant (self):

		# Confirm the file was specified
		if self.common_file:

			# Decompress the gzip file
			with gzip.open(self.common_file, "rt") as common_handle:

				# Loop the common file, record by record
				for record in SeqIO.parse(common_handle, "fasta"):

					# Yield the record as a fasta record
					yield record

def moveFile (file_to_move, out_path):

	# Assing the base filename of the file
	file_basename = os.path.basename(file_to_move)

	# Create the new filename
	new_file_filepath = os.path.join(out_path, file_basename)

	# Move the files
	shutil.move(file_to_move, new_file_filepath)

	# Update the filename
	return new_file_filepath



import os
import sys
import csv
import logging

from kocher_tools.common import readCommonFile
from kocher_tools.database import insertValues, updateValues, retrieveValues, confirmValue

def convertLoc (cursor, table, loc_column, loc_value, cvt_column):
	
	# Retrieve the desired information from the database
	location_data = retrieveValues(cursor, [table], {'IN':{loc_column:[loc_value]}}, [table + '.' +cvt_column])

	# Update log
	logging.info('Location conversion successful')

	# Return the data
	return location_data[0].keys()[0], location_data[0][0]

def addLocFileToDatabase (cursor, table, loc_file):

	# Update log
	logging.info('Uploading location file (%s) to database' % loc_file)

	# Loop the loc file by line
	for header, loc_data in readCommonFile(loc_file):

		# Insert the loc into the database
		insertValues(cursor, table, header, loc_data)

	# Update log
	logging.info('Upload successful')

def updateLocFileToDatabase (cursor, table, select_key, loc_file):

	# Update log
	logging.info('Uploading location file (%s) to database' % loc_file)

	# Loop the loc file by line
	for header, loc_data in readCommonFile(loc_file):

		# Check if the selection key isn't among the headers
		if select_key not in header:
			raise Exception('Selection key (%s) not found. Please check the input file' % select_key)

		# Create an empty string to store the select_value
		select_value = ''

		# Create an empty set dict
		loc_set_dict = {}

		# Create an empty selection dict
		loc_select_dict = {}
		loc_select_dict['IN'] = {}

		# Loop the header ans sample data
		for header_column, loc_value in zip(header, loc_data):

			# Check if the current column is the selection key
			if header_column == select_key:

				# Update the select value
				select_value = loc_value

				# Populate the selection dict
				loc_select_dict['IN'][header_column] = [loc_value]

			else:

				# Populate the selection dict
				loc_set_dict[header_column] = loc_value

		# Check that selected value is present
		if not confirmValue(cursor, table, select_key, select_value):

			# If not, log warning
			logging.warning('Entry (%s) not found, unable to update. Please check input' % select_value)

		# Update the values for the selected value
		updateValues(cursor, table, loc_select_dict, loc_set_dict)

	# Update log
	logging.info('Upload successful')

def readAppLocations (collection_filename):

	with open(collection_filename) as collection_file:

		# Read the data file
		collection_reader = csv.reader(collection_file, delimiter = '\t')

		if sys.version_info[0] == 3:

			# Save the header
			header = next(collection_reader)

		else:
		
			# Save the header
			header = collection_reader.next()

		try:

			# Obtain the index of the GPS data
			gps_index = header.index('GPS')

		except:

			raise Exception('Unable to assign GPS index')

		try:

			# Obtain the index of the site code data
			site_code_index = header.index('Site Code')

		except:

			raise Exception('Unable to assign Site Code index')

		# Create list to hold unique locations
		unique_locations = []

		# Loop the rows of data
		for row in collection_reader:

			# Check if the row has a site code
			if not row[site_code_index]:

				# Check that the row isn't empty
				if len(set(row)) > 1:
					raise Exception('No Site Code found %s' % row)

				# Skip iteration due to empty line
				continue

			# Create a list of the current location
			current_location = [row[site_code_index], row[gps_index]]

			# Check if the current location isnt within the list of unique locations
			if current_location not in unique_locations:

				# Add the current location to the location list
				unique_locations.append(current_location)

		# Loop the unique locations
		for unique_location in unique_locations:

			# Yield the location data
			yield [header[site_code_index], header[gps_index]], unique_location

def addAppLocationsToDatabase (cursor, table, app_file):

	# Update log
	logging.info('Uploading locations from app file (%s) to database' % app_file)

	# Loop the loc file by line
	for header, app_data in readAppLocations(app_file):

		# Insert the loc into the database
		insertValues(cursor, table, header, app_data)

	# Update log
	logging.info('Upload successful')

def updateAppLocationsToDatabase (cursor, table, select_key, app_file):

	# Update log
	logging.info('Uploading locations from app file (%s) to database' % app_file)

	# Loop the loc file by line
	for header, app_data in readAppLocations(app_file):

		# Check if the selection key isn't among the headers
		if select_key not in header:
			raise Exception('Selection key (%s) not found. Please check the input file' % select_key)

		# Create an empty string to store the select_value
		select_value = ''

		# Create an empty set dict
		loc_set_dict = {}

		# Create an empty selection dict
		loc_select_dict = {}
		loc_select_dict['IN'] = {}

		# Loop the header ans sample data
		for header_column, loc_value in zip(header, app_data):

			# Check if the current column is the selection key
			if header_column == select_key:

				# Update the select value
				select_value = loc_value

				# Populate the selection dict
				loc_select_dict['IN'][header_column] = [loc_value]

			else:

				# Populate the selection dict
				loc_set_dict[header_column] = loc_value

		# Check that selected value is present
		if not confirmValue(cursor, table, select_key, select_value):

			# If not, log warning
			logging.warning('Entry (%s) not found, unable to update. Please check input' % select_value)

		# Update the values for the selected value
		updateValues(cursor, table, loc_select_dict, loc_set_dict)

	# Update log
	logging.info('Upload successful')


import os
import sys
import argparse
import copy
import logging
import sqlite3

from collections import defaultdict

from kocher_tools.collection import addAppCollectionToDatabase, updateAppCollectionToDatabase
from kocher_tools.location import addLocFileToDatabase, updateLocFileToDatabase, addAppLocationsToDatabase, updateAppLocationsToDatabase
from kocher_tools.barcode import addSeqFilesToDatabase, updateSeqFilesToDatabase
from kocher_tools.storage import addStorageFileToDatabase, updateStorageFileToDatabase
from kocher_tools.assignment import assignSelectionDict, assignTables
from kocher_tools.database import updateValues
from kocher_tools.config_file import readConfig
from kocher_tools.backup import Backups
from kocher_tools.logger import startLogger, logArgs

def upload_sample_parser ():
	'''
	Argument parser for adding samples

	Raises
	------
	IOError
		If the input, or other specified files do not exist
	'''

	def confirmFile ():
		'''Custom action to confirm file exists'''
		class customAction(argparse.Action):
			def __call__(self, parser, args, value, option_string=None):
				if not os.path.isfile(value):
					raise IOError('%s not found' % value)
				setattr(args, self.dest, value)
		return customAction

	def confirmFileList ():
		'''Custom action to confirm file exists in list'''
		class customAction(argparse.Action):
			def __call__(self, parser, args, values, option_string=None):
				# Loop the list
				for value in values:
					# Check if the file exists
					if not os.path.isfile(value):
						raise IOError('%s not found' % value)
				if not getattr(args, self.dest):
					setattr(args, self.dest, values)
				else:
					getattr(args, self.dest).extend(values)
		return customAction

	def confirmLinkedFiles ():
		'''Custom action to confirm file exists in list'''
		class customAction(argparse.Action):
			def __call__(self, parser, args, values, option_string=None):
				# Loop the list
				for value in values:
					# Check if the file exists
					if not os.path.isfile(value):
						raise IOError('%s not found' % value)
				if not getattr(args, self.dest):
					setattr(args, self.dest, [values])
				else:
					getattr(args, self.dest).append(values)
		return customAction

	def updateDict ():
		'''Custom action to add items to a list'''
		class customAction(argparse.Action):
			def __call__(self, parser, args, value_list, option_string=None):

				# Clean up any commas
				value_list = [item.strip(',') for item in value_list]

				if not getattr(args, self.dest):
					# Create the dict
					value_dict = {}
					# Populate the dict
					value_dict[value_list[0]] = value_list[1]
					setattr(args, self.dest, value_dict)
				else:
					getattr(args, self.dest)[value_list[0]] = value_list[1]
		return customAction

	def selectionList ():
		'''Custom action to add items to a list'''
		class customAction(argparse.Action):
			def __call__(self, parser, args, value_list, option_string=None):

				# Clean up any commas
				value_list = [item.strip(',') for item in value_list]

				if not getattr(args, self.dest):
					setattr(args, self.dest, value_list)
				else:
					getattr(args, self.dest).extend(value_list)
		return customAction

	def selectionDict ():
		'''Custom action to add items to a list'''
		class customAction(argparse.Action):
			def __call__(self, parser, args, value_list, option_string=None):

				# Clean up any commas
				value_list = [item.strip(',') for item in value_list]

				if not getattr(args, self.dest):
					# Create the dict
					value_dict = defaultdict(list)
					# Populate the dict
					value_dict[value_list[0]].append(value_list[1])
					setattr(args, self.dest, value_dict)
				else:
					getattr(args, self.dest)[value_list[0]].append(value_list[1])
		return customAction

	def metavar_list (var_list):
		'''Create a formmated metavar list for the help output'''
		return '{' + ', '.join(var_list) + '}'

	upload_parser = argparse.ArgumentParser(formatter_class = argparse.ArgumentDefaultsHelpFormatter)

	# Input arguments
	upload_parser.add_argument('--app-file', dest = 'app_files', help = "Defines the collection app filename", type = str, nargs = '+', action = confirmFileList())
	upload_methods = ('Collection', 'Location', 'Both')
	upload_parser.add_argument('--app-upload-method', metavar = metavar_list(upload_methods), help = 'Upload method for collection app files', choices = upload_methods, default = upload_methods[0])
	upload_parser.add_argument('--barcode-file-pair', dest = 'barcode_files', metavar = ('blast_file', 'fasta_file'), help = "Defines the BLAST and fasta output filename", type = str, nargs = 2, action = confirmLinkedFiles())
	upload_parser.add_argument('--barcode-all-files', dest = 'barcode_files', metavar = ('blast_file', 'fasta_file', 'failed_file'), help = "Defines the BLAST, fasta, and failed output filename", type = str, nargs = 3, action = confirmLinkedFiles())
	upload_parser.add_argument('--location-file', dest = 'loc_files', help = 'Defines the filename of a location file', type = str, nargs = '+', action = confirmFileList())
	upload_parser.add_argument('--storage-file', dest = 'storage_files', help = 'Defines the filename of a storage file ', type = str, nargs = '+', action = confirmFileList())
	
	# Selection arguments
	upload_parser.add_argument('--include-ID', help = 'ID to include in database updates', type = str, nargs = '+', action = selectionList())
	upload_parser.add_argument('--exclude-ID', help = 'ID to exclude in database updates', type = str, nargs = '+', action = selectionList())
	upload_parser.add_argument('--include-species', help = 'Species to include in database updates', type = str, nargs = '+', action = selectionList())
	upload_parser.add_argument('--exclude-species', help = 'Species to exclude in database updates', type = str, nargs = '+', action = selectionList())
	upload_parser.add_argument('--include-genus', help = 'Species to include in database updates', type = str, nargs = '+', action = selectionList())
	upload_parser.add_argument('--exclude-genus', help = 'Species to exclude in database updates', type = str, nargs = '+', action = selectionList())
	upload_parser.add_argument('--include-nests', help = 'Include samples from nests in database retrievals', action = 'store_true')
	upload_parser.add_argument('--exclude-nests', help = 'Exclude samples from nests in database retrievals', action = 'store_true')
	upload_parser.add_argument('--include', metavar = ('column', 'value'), help = 'Column/value pair to include in database updates', type = str, nargs = 2, action = selectionDict())
	upload_parser.add_argument('--exclude', metavar = ('column', 'value'), help = 'Column/value pair to exclude in database updates', type = str, nargs = 2, action = selectionDict())

	# Update arguments
	upload_parser.add_argument('--update-with-file', help = 'Update database using with the entries within a table', action='store_true')
	upload_parser.add_argument('--update', metavar = ('column', 'value'), help = 'Column/value pair to update', type = str, nargs = 2, action = updateDict())

	# Output arguments
	upload_parser.add_argument('--out-log', help = 'Filename of the log file', type = str, default = 'upload_samples.log')
	upload_parser.add_argument('--log-stdout', help = 'Direct logging to stdout', action = 'store_true')

	# Database arguments
	upload_parser.add_argument('--sqlite-db', help = 'SQLite database filename', type = str, required = True, action = confirmFile())
	upload_parser.add_argument('--yaml', help = 'Database YAML config file', type = str, required = True, action = confirmFile())
	upload_parser.add_argument('--backup', help = 'Will force the creation of a backup', action='store_true')
	upload_parser.add_argument('--no-backup', help = 'Will skip backup procedure', action='store_true')

	return upload_parser.parse_args()

def canUpdate (update_dict):

	# Create list of columns that shouldn't be altered
	locked_list = ['Site Code', 'Unique ID']

	# loop the update dict by columns
	for column in update_dict.keys():

		# Check if the column is within the locked list
		if column.strip().replace('"','') in locked_list:

			return False

	return True

def main():
			
	# Assign arguments
	upload_args = upload_sample_parser()

	# Check if log should be sent to stdout
	if upload_args.log_stdout:

		# Start logging to stdout for this run
		startLogger()

	else:

		# Start a log file for this run
		startLogger(log_filename = upload_args.out_log)

	# Read in the database config file
	db_config_data = readConfig(upload_args.yaml)
	
	# Log the arguments used
	logArgs(upload_args)

	# Check if the backup process should be skipped
	if upload_args.no_backup:

		logging.info('Skipping backup procedure')

	else:

		# Load the current backups
		current_backups = Backups(out_dir = db_config_data.backup_out_dir, limit = db_config_data.backup_limit, update_freq = db_config_data.backup_update_freq)

		# Check if a backup is required
		if current_backups.backupNeeded() or upload_args.backup:

			# Backup the database
			current_backups.newBackup(upload_args.sqlite_db)

	# Connect to the sqlite database
	sqlite_connection = sqlite3.connect(upload_args.sqlite_db)

	# Setup SQLite to reture the rows as dict with columns
	sqlite_connection.row_factory = sqlite3.Row

	# Create the cursor
	cursor = sqlite_connection.cursor()

	# Check if a collection app file has been specified
	if upload_args.app_files:

		# Loop the collection app files
		for app_file in upload_args.app_files:

			# Check if a selection key has been specified
			if upload_args.update_with_file:

				# Check if the Collection should be updated
				if upload_args.app_upload_method in ['Collection', 'Both']:

					# Update the database with the file
					updateAppCollectionToDatabase(cursor, 'Collection', 'Unique ID', app_file)

				# Check if the Collection should be updated
				if upload_args.app_upload_method in ['Location', 'Both']:

					# Update the database with the file
					updateAppLocationsToDatabase(cursor, 'Locations', 'Site Code', app_file)

			else:

				# Check if the Collection should be added
				if upload_args.app_upload_method in ['Collection', 'Both']:

					# Add file to the database
					addAppCollectionToDatabase(cursor, 'Collection', app_file)

				# Check if the Collection should be updated
				if upload_args.app_upload_method in ['Location', 'Both']:

					# Add file to the database
					addAppLocationsToDatabase(cursor, 'Locations', app_file)

	# Check if barcode files have been specified
	if upload_args.barcode_files:

		# Loop the collection app files
		for barcode_files in upload_args.barcode_files:

			# Check if only a pair of files was passed
			if len(barcode_files) == 2: 

				# Assign the blast and fasta file
				blast_file, fasta_file = barcode_files

				# Set the failed file to None
				failed_file = None

			# Check if all files was passed
			elif len(barcode_files) == 3: 

				# Assign the blast and fasta file
				blast_file, fasta_file, failed_file = barcode_files

			# Check if a selection key has been specified
			if upload_args.update_with_file:

				# Update the database with the file
				updateSeqFilesToDatabase(cursor, 'Sequencing', 'Sequence ID', blast_file, fasta_file, failed_file, 'Storage', 'Storage ID')

			else:

				# Add file to the database
				addSeqFilesToDatabase(cursor, 'Sequencing', blast_file, fasta_file, failed_file, 'Storage', 'Storage ID')

	# Check if a location file has been specified
	if upload_args.loc_files:

		# Loop the location files
		for loc_file in upload_args.loc_files:

			# Check if a selection key has been specified
			if upload_args.update_with_file:

				# Update the database with the file
				updateLocFileToDatabase(cursor, 'Locations', 'Site Code', loc_file)

			else:

				# Add file to the database
				addLocFileToDatabase(cursor, 'Locations', loc_file)

	# Check if a storage file has been specified
	if upload_args.storage_files:

		# Loop the storage files
		for storage_file in upload_args.storage_files:

			# Check if a selection key has been specified
			if upload_args.update_with_file:

				# Update the database with the file
				updateStorageFileToDatabase(cursor, 'Storage', 'Storage ID', storage_file)

			else:

				# Add file to the database
				addStorageFileToDatabase(cursor, 'Storage', storage_file)

	# Check if update data has been specified
	if upload_args.update:

		# Check if the update is impossible
		if not canUpdate(upload_args.update):
			raise Exception('Not allowed to alter: %s' % column)

		# Assign the tables that need to be updated
		selection_tables = assignTables(db_config_data, **vars(upload_args))

		# Assign a defaultdict with all the selection information
		selection_dict = assignSelectionDict(db_config_data, **vars(upload_args))

		# Create a list of the tables to update
		tables_to_update = db_config_data.returnTables(list(upload_args.update.keys()))
		
		# Loop the tables that need to be updated 
		for table_to_update in tables_to_update:

			# Create the update statement dict for the current table
			update_statement_dict = db_config_data.returnColumnDict(upload_args.update, table_to_update)

			# Check if the current table is within the tables to join
			if table_to_update in selection_tables and len(selection_tables) == 1:

				# Update the specified data 
				updateValues(cursor, table_to_update, selection_dict, update_statement_dict)

			# Run the expanded command if there are tables to join
			else:

				# Assign the key for the table to update
				join_by_column = db_config_data[table_to_update].join_by_key

				# Create a copy of the selection tables
				tables_to_join = copy.copy(selection_tables)

				# Add the table that is going to be updated
				tables_to_join.append(table_to_update)

				# Remove any duplicates
				tables_to_join = list(set(tables_to_join))

				# Assign the tables and keys that need to be joined
				join_by_names, join_by_columns = db_config_data.returnJoinLists(tables_to_join)

				# Update the specified data 
				updateValues(cursor, table_to_update, selection_dict, update_statement_dict, update_table_column = join_by_column, tables_to_join = join_by_names, join_table_columns = join_by_columns)

	# Commit any changes
	sqlite_connection.commit()

	# Close the connection
	cursor.close()

if __name__ == "__main__":
	main()


#!/usr/bin/env python
import os
import sys
import argparse
import logging
import getpass
import sqlite3

from kocher_tools.logger import startLogger, logArgs
from kocher_tools.config_file import ConfigDB
from kocher_tools.database import *

def confirmFile ():
	'''Custom action to confirm file exists'''
	class customAction(argparse.Action):
		def __call__(self, parser, args, value, option_string=None):
			if not os.path.isfile(value):
				raise IOError('%s not found' % value)
			setattr(args, self.dest, value)
	return customAction

# Define args
db_parser = argparse.ArgumentParser(formatter_class = argparse.ArgumentDefaultsHelpFormatter)
db_parser.add_argument('--yaml', help = "Database YAML config file", type = str, required = True, action = confirmFile())
db_parser.add_argument('--out-log', help = 'Filename of the log file', type = str, default = 'create_database.log')
db_args = db_parser.parse_args()

# Start a log file for this run
startLogger(log_filename = db_args.out_log)

# Log the arguments used
logArgs(db_args)

# Read in the YAML config file
db_config_data = ConfigDB.readConfig(db_args.yaml)

# Get the password, if not found
if db_config_data.passwd_required and not db_config_data.passwd:
	db_config_data.passwd = getpass.getpass("Password: ")
	
# Create the tables
sql_engine = createEngineFromConfig(db_config_data)
createAllFromConfig(db_config_data, sql_engine)